"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const aff_1 = require("./aff");
const fs_extra_1 = require("fs-extra");
const iconv_lite_1 = require("iconv-lite");
const fixRegex = {
    'SFX': { m: /$/, r: '$' },
    'PFX': { m: /^/, r: '^' },
};
const emptyZeroRegex = /^0$/;
const yesRegex = /[yY]/;
const spaceRegex = /\s+/;
const commentRegex = /(?:^\s*#.*)|(?:\s+#.*)/;
const UTF8 = 'UTF-8';
function convEntry(fieldValue, _, args) {
    if (fieldValue === undefined) {
        return [];
    }
    fieldValue.push({ from: args[0], to: args[1] });
    return fieldValue;
}
function afEntry(fieldValue, _, args) {
    if (fieldValue === undefined) {
        return [''];
    }
    fieldValue.push(args[0]);
    return fieldValue;
}
function simpleTable(fieldValue, _, args) {
    if (fieldValue === undefined) {
        const [count, ...extraValues] = args;
        const extra = extraValues.length ? extraValues : undefined;
        return { count, extra, values: [] };
    }
    fieldValue.values.push(args);
    return fieldValue;
}
const regExpStartsWithPlus = /^\+/;
function tablePfxOrSfx(fieldValue, _, args, type) {
    /*
    Fields of an affix rules:
    (0) Option name
    (1) Flag
    (2) stripping characters from beginning (at prefix rules) or end (at suffix rules) of the word
    (3) affix (optionally with flags of continuation classes, separated by a slash)
    (4) condition.
        Zero stripping or affix are indicated by zero. Zero condition is indicated by dot.
        Condition is a simplified, regular expression-like pattern, which must be met before the affix can be applied.
        (Dot signs an arbitrary character. Characters in braces sign an arbitrary character from the character subset.
        Dash hasn't got special meaning, but circumflex (^) next the first brace sets the complementer character set.)
    (5) Optional morphological fields separated by spaces or tabulators.
     */
    const posCondition = 2;
    if (fieldValue === undefined) {
        fieldValue = new Map();
    }
    const [subField, ...subValues] = args;
    if (!fieldValue.has(subField)) {
        const id = subField;
        const [combinable, count, ...extra] = subValues;
        fieldValue.set(subField, {
            id,
            type,
            combinable: !!combinable.match(yesRegex),
            count,
            extra,
            substitutionSets: new Map()
        });
        return fieldValue;
    }
    if (subValues.length < 2) {
        console.log(`Affix rule missing values: ${args.join(' ')}`);
        return;
    }
    if (regExpStartsWithPlus.test(subValues[posCondition] || '')) {
        // sometimes the condition is left off, but there are morphological fields
        // so we need to inject a '.'.
        subValues.splice(posCondition, 0, '.');
    }
    const fixRuleSet = fieldValue.get(subField);
    const substitutionSets = fixRuleSet.substitutionSets;
    const [removeValue, attach, ruleAsString = '.', ...extraValues] = subValues;
    const [attachText, attachRules] = attach.split('/', 2);
    const extra = extraValues.length ? extraValues : undefined;
    const remove = removeValue.replace(emptyZeroRegex, '');
    const insertText = attachText.replace(emptyZeroRegex, '');
    const fixUp = fixRegex[type];
    const replace = new RegExp(remove.replace(fixUp.m, fixUp.r));
    if (!substitutionSets.has(ruleAsString)) {
        const match = new RegExp(ruleAsString.replace(fixUp.m, fixUp.r));
        substitutionSets.set(ruleAsString, {
            match,
            substitutions: [],
        });
    }
    const substitutionSet = substitutionSets.get(ruleAsString);
    substitutionSet.substitutions.push({ remove, replace, attach: insertText, attachRules, extra });
    return fieldValue;
}
function asPfx(fieldValue, field, args) {
    return tablePfxOrSfx(fieldValue, field, args, 'PFX');
}
function asSfx(fieldValue, field, args) {
    return tablePfxOrSfx(fieldValue, field, args, 'SFX');
}
function asString(_fieldValue, _field, args) {
    return args[0];
}
function asBoolean(_fieldValue, _field, args) {
    const [value = '1'] = args;
    const iValue = parseInt(value);
    return !!iValue;
}
function asNumber(_fieldValue, _field, args) {
    const [value = '0'] = args;
    return parseInt(value);
}
/*
cspell:ignore COMPOUNDBEGIN COMPOUNDEND COMPOUNDMIDDLE COMPOUNDMIN COMPOUNDPERMITFLAG COMPOUNDRULE FORBIDDENWORD KEEPCASE
cspell:ignore MAXDIFF NEEDAFFIX WORDCHARS
*/
const affTableField = {
    AF: afEntry,
    BREAK: asNumber,
    CHECKCOMPOUNDCASE: asBoolean,
    CHECKCOMPOUNDDUP: asBoolean,
    CHECKCOMPOUNDPATTERN: simpleTable,
    CHECKCOMPOUNDREP: asBoolean,
    COMPOUNDBEGIN: asString,
    COMPOUNDEND: asString,
    COMPOUNDMIDDLE: asString,
    COMPOUNDMIN: asNumber,
    COMPOUNDPERMITFLAG: asString,
    COMPOUNDRULE: simpleTable,
    FLAG: asString,
    FORBIDDENWORD: asString,
    FORCEUCASE: asString,
    ICONV: convEntry,
    KEEPCASE: asString,
    KEY: asString,
    MAP: simpleTable,
    MAXCPDSUGS: asNumber,
    MAXDIFF: asNumber,
    NEEDAFFIX: asString,
    NOSPLITSUGS: asBoolean,
    NOSUGGEST: asString,
    OCONV: convEntry,
    ONLYINCOMPOUND: asString,
    ONLYMAXDIFF: asBoolean,
    PFX: asPfx,
    REP: simpleTable,
    SET: asString,
    SFX: asSfx,
    TRY: asString,
    WARN: asString,
    WORDCHARS: asString,
};
function parseAffFile(filename, encoding = UTF8) {
    return __awaiter(this, void 0, void 0, function* () {
        const buffer = yield fs_extra_1.readFile(filename);
        const file = iconv_lite_1.decode(buffer, encoding);
        const affInfo = parseAff(file, encoding);
        if (affInfo.SET && affInfo.SET.toLowerCase() !== encoding.toLowerCase()) {
            return parseAff(iconv_lite_1.decode(buffer, affInfo.SET.toLowerCase()), affInfo.SET);
        }
        return affInfo;
    });
}
exports.parseAffFile = parseAffFile;
function parseAff(affFileContent, _encoding = UTF8) {
    const lines = affFileContent.split(/\r?\n/g);
    return lines
        .map(line => line.trimLeft())
        .map(line => line.replace(commentRegex, ''))
        .filter(line => line.trim() !== '')
        .map(line => line.split(spaceRegex))
        .reduce((aff, line) => {
        const [field, ...args] = line;
        const fn = affTableField[field];
        if (fn) {
            aff[field] = fn(aff[field], field, args);
        }
        else {
            aff[field] = args;
        }
        return aff;
    }, {});
}
exports.parseAff = parseAff;
function parseAffFileToAff(filename, encoding) {
    return parseAffFile(filename, encoding)
        .then(affInfo => new aff_1.Aff(affInfo));
}
exports.parseAffFileToAff = parseAffFileToAff;
//# sourceMappingURL=affReader.js.map